import csv

locus = 'IGH'
input_file = 'reference_sets/cirelli.csv'
output_file = 'reference_sets/cirelli_IGH.fasta'

with open(input_file, 'r') as fi, open(output_file, 'w') as fo:
    reader = csv.reader(fi, dialect='excel')
    for row in reader:
        if locus in row[0]:
            fo.write('>%s\n%s\n' % (row[0], row[1]))




#!python
# Annotate a reference assembly given a starting germline set

# Copyright (c) 2021 William Lees

# This source code, and any executable file compiled or derived from it, is governed by the European Union Public License v. 1.2,
# the English version of which is available here: https://perma.cc/DK5U-NDVE

import argparse
import os.path
import subprocess
from receptor_utils import simple_bio_seq as simple
import pathlib
import glob
try:
    from slugify import slugify
except:
    from digger.slugify import slugify


def get_parser():
    parser = argparse.ArgumentParser(description='Find functional and nonfunctional genes in a single assembly sequence')
    parser.add_argument('assembly_file', help='assembly sequence to search')
    parser.add_argument('-species', help='use motifs for the specified species provided with the package')
    parser.add_argument('-motif_dir', help='pathname to directory containing motif probability files')
    parser.add_argument('-locus', help='locus (default is IGH)')
    parser.add_argument('-v_ref', help='set of V reference genes to use as starting point for search')
    parser.add_argument('-d_ref', help='set of D reference genes to use as starting point for search')
    parser.add_argument('-j_ref', help='set of J reference genes to use as starting point for search')
    parser.add_argument('-v_ref_gapped', help='IMGT-gapped v-reference set used to determine alignment of novel sequences')
    parser.add_argument('-ref_comp', help='ungapped reference set(s) to compare to: name and reference file separated by comma eg mouse,mouse.fasta (may be repeated multiple times)', action='append')
    parser.add_argument('-sense', help='sense in which to read the assembly (forward or reverse) (if omitted will select automatically)')
    parser.add_argument('-keepwf', help='keep working files after processing has completed', action='store_true')
    parser.add_argument('output_file', help='output file (csv)')
    return parser


# A function to remove intermediate files

def remove_working_files():
    for fn in ['full_germline_set.fasta', 'assembly.fasta', 'assembly_rc.fasta', 'blast_results_assembly.csv']:
        if os.path.isfile(fn):
            os.remove(fn)

    for fn in glob.glob('assembly*.out'):
        if os.path.isfile(fn):
            os.remove(fn)

    for ext in ['.ndb', '.nhr', '.nin', '.njs', '.not', '.nsq', '.ntf', '.nto']:
        for fn in glob.glob('*' + ext):
            if os.path.isfile(fn):
                os.remove(fn)


def main():
    args = get_parser().parse_args()

    cwd = pathlib.Path().resolve()
    remove_working_files()

    if os.path.isfile(args.output_file):
        os.remove(args.output_file)

    # check arguments

    for fn in [args.assembly_file, args.v_ref]:
        if not os.path.isfile(fn):
            print(f"{fn} - file not found")
            exit(1)

    if not args.motif_dir and not args.species:
        print('Error - please specify either -motif_dir or -species')
        quit()

    if args.motif_dir and args.species:
        print('Error - please specify either -motif_dir or -species, not both')
        quit()


    if args.motif_dir and not os.path.isdir(args.motif_dir):
        print(f"{args.motif_dir} - directory not found")
        exit(1)

    full_germline_set = {}
    for fn in [args.v_ref, args.d_ref, args.j_ref]:
        if fn and not os.path.isfile(fn):
            print(f"{fn} - file not found")
            exit(1)

        if fn:
            full_germline_set |= simple.read_fasta(fn)

    simple.write_fasta('full_germline_set.fasta', full_germline_set)

    if args.ref_comp:
        for ref_arg in args.ref_comp:
            ref_name, fn = ref_arg.split(',')
            if not os.path.isfile(fn):
                print(f"{fn} - file not found")
                exit(1)

    if not(args.v_ref or args.d_ref or args.j_ref):
        print('Please specify at least one starting point reference set')
        exit(1)

    assembly_contents = simple.read_fasta(args.assembly_file)

    if len(assembly_contents) != 1:
        print('The assembly file must contain exactly one sequence.')
        exit(1)

    simple.write_fasta('assembly.fasta', {list(assembly_contents.keys())[0]: list(assembly_contents.values())[0]})
    assembly_name = slugify(list(assembly_contents.keys())[0])
    assembly_length = len(list(assembly_contents.values())[0])

    locus = 'IGH'

    if args.locus:
        if args.locus not in ['IGH', 'IGK', 'IGL', 'TRA', 'TRB', 'TRD', 'TRG']:
            print(f"locus must be one of {','.join(['IGH', 'IGK', 'IGL', 'TRA', 'TRB', 'TRD', 'TRG'])}")
            exit(1)
        locus = args.locus

    # Make blast databases

    for fn in [args.v_ref, args.d_ref, args.j_ref]:
        if fn:
            cmd = [
                    'makeblastdb',
                    '-in', f'{fn}',
                    '-dbtype', 'nucl',
            ]
            print(f"\n-- executing {' '.join(cmd)}\n")
            process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            #if not os.path.isfile(fn+'.ndb'):
            #    print('Blast database not created - quitting')
            #    exit(1)

    # Run blast and process output

    first_time = True

    for fn, gene_type in [(args.v_ref, 'V'), (args.d_ref, 'D'), (args.j_ref, 'J')]:
        if fn:
            if gene_type == 'J' or gene_type == 'D':
                word_size = '7'
            else:
                word_size = '11'

            if gene_type == 'D':
                evalue = '100'
            else:
                evalue = '10'

            cmd = [
                'blastn',
                '-db', f'{fn}',
                '-query', 'assembly.fasta',
                '-out', f"assembly_{gene_type}.out",
                '-outfmt', '7',
                '-gapopen', '5',
                '-gapextend', '5',
                '-penalty', '-1',
                '-word_size', word_size,
                '-evalue', evalue,
            ]
            print(f"\n-- executing {' '.join(cmd)}\n")
            process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            if not os.path.isfile(f"assembly_{gene_type}.out"):
                print('Blast output not created - quitting')
                exit(1)

            cmd = [
                'blastresults_to_csv',
                f"assembly_{gene_type}.out",
                'blast_results_',
            ]

            if not first_time:
                cmd.append('-a')

            first_time = False

            print(f"\n-- executing {' '.join(cmd)}\n")
            process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    if not os.path.isfile(f'blast_results_{assembly_name}.csv'):
        print('Blast results not converted to csv - quitting')
        exit(1)

    # Run find_alignments

    if args.motif_dir:
        cmd = [
            'find_alignments',
            'full_germline_set.fasta',
            'assembly.fasta',
            f'blast_results_{assembly_name}.csv',
            '-motif_dir',  f'{args.motif_dir}',
            '-locus', locus,
            ]
    else:
        cmd = [
            'find_alignments',
            'full_germline_set.fasta',
            'assembly.fasta',
            f'blast_results_{assembly_name}.csv',
            '-species', f'{args.species}',
            '-locus', locus,
            ]

    if args.ref_comp:
        for ref_arg in args.ref_comp:
            cmd.extend(['-ref', ref_arg])

    if args.sense:
        cmd.extend(['-sense', args.sense])

    if args.v_ref_gapped:
        cmd.extend(['-align', args.v_ref_gapped])

    cmd.append(args.output_file)

    print(f"\n-- executing {' '.join(cmd)}\n")
    process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    print(process.stdout.decode("utf-8"))
    reverse_sense = 'Using reverse' in process.stdout.decode("utf-8")

    if not os.path.isfile(args.output_file):
        print(f'Result file {args.output_file} was not produced by find_alignments.py')
        exit(1)

    if not args.keepwf:
        remove_working_files()
    exit(0)


if __name__ == "__main__":
    main()





# Process a single sequence, given a guide gene to look for
# The sequence may be nominated as a fasta file, or by genbank accession number, or by a list of genbank accession numbers in a CSV file
import argparse
from receptor_utils import simple_bio_seq as simple
import csv

try:
    from dig_utils import set_motif_params, read_motifs, read_vs, process_sequence, process_output, get_genbank_sequence
except:
    from digger.dig_utils import set_motif_params, read_motifs, read_vs, process_sequence, process_output, get_genbank_sequence



def get_parser():
    parser = argparse.ArgumentParser(description='Annotate a genomic sequence representing the nominated receptor gene')
    subparsers = parser.add_subparsers(required=True)

    parser_fasta = subparsers.add_parser('fasta', help='Annotate a single genomic sequence in a FASTA file')
    parser_fasta.add_argument('target', help='Name of nominated sequence in reference set')
    parser_fasta.add_argument('germline_file', help='ungapped reference set containing the nominated sequence (FASTA)')
    parser_fasta.add_argument('query_file', help='file containing the sequence to annotate (FASTA)')
    parser_fasta.add_argument('-align', help='gapped reference set to use for V gene alignments (required for V gene analysis')
    parser_fasta.add_argument('-species', help='use motifs for the specified species provided with the package')
    parser_fasta.add_argument('-motif_dir', help='use motif probability files present in the specified directory')
    parser_fasta.add_argument('-out_file', help='output file (CSV)')
    parser_fasta.add_argument('-debug', help='produce parsing_errors file with debug information', action='store_true')
    parser_fasta.set_defaults(func=main_fasta)

    parser_single = subparsers.add_parser('single', help='Annotate a single sequence given its genbank accession number')
    parser_single.add_argument('target', help='Name of nominated sequence')
    parser_single.add_argument('germline_file', help='ungapped reference set containing the nominated sequence (FASTA)')
    parser_single.add_argument('genbank_acc', help='genbank accession number of the sequence to annotate')
    parser_single.add_argument('-align', help='gapped reference set to use for V gene alignments (required for V gene analysis')
    parser_single.add_argument('-species', help='use motifs for the specified species provided with the package')
    parser_single.add_argument('-motif_dir', help='use motif probability files present in the specified directory')
    parser_single.add_argument('-out_file', help='output file (CSV)')
    parser_single.set_defaults(func=main_single)

    parser_multi = subparsers.add_parser('multi', help='Read allele names and corresponding genbank accession numbers from a CSV file')
    parser_multi.add_argument('locus', help='Locus of nominated sequences')
    parser_multi.add_argument('germline_file', help='ungapped reference set containing the nominated sequence (FASTA)')
    parser_multi.add_argument('query_file', help='File containing list of targets and associated genbank accession numbers (CSV)')
    parser_multi.add_argument('-align', help='gapped reference set to use for V gene alignments (required for V gene analysis')
    parser_multi.add_argument('-species', help='use motifs for the specified species provided with the package')
    parser_multi.add_argument('-motif_dir', help='use motif probability files present in the specified directory')
    parser_multi.add_argument('-out_file', help='output file (CSV)')
    parser_multi.set_defaults(func=main_multi)

    parser_multi = subparsers.add_parser('multi_seq', help='Read allele names and genomic sequences from a CSV file')
    parser_multi.add_argument('locus', help='Locus of nominated sequences')
    parser_multi.add_argument('germline_file', help='ungapped reference set containing the nominated sequence (FASTA)')
    parser_multi.add_argument('query_file', help='File containing list of targets and genomic sequences (CSV)')
    parser_multi.add_argument('-align', help='gapped reference set to use for V gene alignments (required for V gene analysis')
    parser_multi.add_argument('-species', help='use motifs for the specified species provided with the package')
    parser_multi.add_argument('-motif_dir', help='use motif probability files present in the specified directory')
    parser_multi.add_argument('-out_file', help='output file (CSV)')
    parser_multi.set_defaults(func=main_multi_seq)

    return parser


def main():
    args = get_parser().parse_args()
    args.func(args)

def main_fasta(args):
    germlines = simple.read_fasta(args.germline_file)
    target = args.target
    locus = target[:3]
    motif_params = set_motif_params(locus)
    conserved_motif_seqs, motifs = read_motifs(args, locus)
    v_gapped_ref, v_ungapped_ref = read_vs(args)

    assembly = simple.read_single_fasta(args.query_file)
    patch = ''

    genbank_acc = 'NA'
    row = process_sequence(assembly, genbank_acc, patch, target, germlines, v_gapped_ref, v_ungapped_ref, motifs, conserved_motif_seqs, motif_params)

    process_output(args, [row])


def main_single(args):
    germlines = simple.read_fasta(args.germline_file)
    target = args.target
    locus = target[:3]
    motif_params = set_motif_params(locus)
    conserved_motif_seqs, motifs = read_motifs(args, locus)
    v_gapped_ref, v_ungapped_ref = read_vs(args)

    if target not in germlines:
        print('Target sequence not found in reference set')
        return

    genbank_res = get_genbank_sequence(args.genbank_acc)

    if args.genbank_acc in genbank_res:
        genbank_res = genbank_res[args.genbank_acc]
        row = process_sequence(genbank_res['seq'], genbank_res['acc'], genbank_res['patch'], target, germlines, v_gapped_ref, v_ungapped_ref, motifs, conserved_motif_seqs, motif_params)
        process_output(args, [row])
    else:
        print('Accession number not found in Genbank')



def main_multi(args):
    multi_driver(args, True)

def main_multi_seq(args):
    multi_driver(args, False)

def multi_driver(args, genomic):
    germlines = simple.read_fasta(args.germline_file)
    locus = args.locus
    motif_params = set_motif_params(locus)
    conserved_motif_seqs, motifs = read_motifs(args, locus)
    v_gapped_ref, v_ungapped_ref = read_vs(args)

    query_list = {}
    if genomic:
        wanted_accs = []
        with open(args.query_file, 'r') as f:
            reader = csv.reader(f)
            for row in reader:
                if row[0] not in query_list:
                    query_list[row[0]] = []
                query_list[row[0]].append({'accession': row[1]})
                if row[1] not in wanted_accs:
                    wanted_accs.append(row[1])

        # Fetch the sequences from genbank and process them
        genbank_res = get_genbank_sequence(','.join(wanted_accs))
        accession_results = {acc: row for acc, row in genbank_res.items()}

        # add sequences to query_list
        for allele in query_list:
            for row in query_list[allele]:
                if row['accession'] not in accession_results:
                    print(f'No sequence found for accession number {row["accession"]}')
                else:
                    row['seq'] = accession_results[row['accession']]['seq']
                    row['patch'] = accession_results[row['accession']]['patch']
    else:
        for row in simple.read_csv(args.query_file):
            if row['allele'] not in query_list:
                query_list[row['allele']] = []
            rec = {'seq': row['sequence'], 'accession': row['accession'], 'patch': ''}
            if 'assembly_offset' in row and 'sense' and 'assembly_length' in row:
                rec['assembly_offset'] = row['assembly_offset']
                rec['sense'] = row['sense']
                rec['assembly_length'] = row['assembly_length']
            query_list[row['allele']].append(rec)

    # Process the sequences

    rows = []
    for target_allele, reqs in query_list.items():
        if target_allele in germlines:
            for req in reqs:
                if 'seq' in req and req['seq']:
                    print(f'Processing {target_allele} {req["accession"]}')
                    row = process_sequence(req['seq'], req['accession'], req['patch'], target_allele, germlines, v_gapped_ref, v_ungapped_ref, motifs, conserved_motif_seqs, motif_params)
                    rows.append(row)

                    if row['start'] != 'NA' and 'assembly_offset' in req:
                        row['sense'] = req['sense']
                        if req['sense'] == '+':
                            for el in ['start', 'end', 'gene_start', 'gene_end', 'l_part1_start', 'l_part1_end', 'l_part2_start', 'l_part2_end', '3_rss_start', '3_rss_end', '5_rss_start', '5_rss_end']:
                                if el in row:
                                    row[el] += int(req['assembly_offset'])
                        else:
                            for el in ['start', 'end', 'gene_start', 'gene_end', 'l_part1_start', 'l_part1_end', 'l_part2_start', 'l_part2_end', '3_rss_start', '3_rss_end', '5_rss_start', '5_rss_end']:
                                if el in row:
                                    row[el] = int(req['assembly_length']) - row[el] + 1

                        for el in list(row.keys()):
                            if 'start' in el or 'end' in el and 'rev' not in el:
                                row[el + '_rev'] = int(req['assembly_length']) - row[el] + 1

        else:
            print(f'No germline sequence found for {target_allele}')

    process_output(args, rows)


if __name__ == "__main__":
    main()



# this uses the representation at http://www.imgt.org/ligmdb/view?format=IMGT&id=IMGT000064

# pull out the V-regions, functional/pseudo and locs

import csv
import os

from Bio.Align import MultipleSeqAlignment, AlignInfo
from Bio.SeqRecord import SeqRecord
from collections import Counter
import argparse
import urllib.request
import html
from receptor_utils import simple_bio_seq as simple

assembly = None

def get_parser():
    parser = argparse.ArgumentParser(description='Given a set of IMGT annotations, build a CSV file containing gene names and co-ordinates')
    parser.add_argument('imgt_url', help='URL of IMGT annotation, e.g. http://www.imgt.org/ligmdb/view?format=IMGT&id=IMGT000064, or name of text file containing its contents')
    parser.add_argument('outfile', help='Output file (CSV)')
    parser.add_argument('locus', help='one of IGH, IGK, IGL, TRA, TRB, TRD, TRG')
    parser.add_argument('--save_download', help='Save contents of annotation to specified file')
    parser.add_argument('--save_sequence', help='Save sequence to specified file')
    parser.add_argument('--save_imgt_annots', help='Save IMGT annotations to specified file')
    return parser


def find_range(line):
    if '..' not in line:
        return (None, None)


    (start, end) = line[25:].split('..')
    start = int(start.replace('<', '').replace('>', ''))
    end = int(end.replace('\n', '').replace('<', '').replace('>', ''))
    return (start, end)

def find_seq(line, extra=0):
    sense = 'forward'
    complement = 'complement' in line
    if complement:
        line = line.replace('complement(','').replace(')', '')
        sense = 'reverse'

    (start, end) = find_range(line)

    if start is None:
        return None

    seq = assembly[start-1:end+extra]

    if complement:
        seq = simple.reverse_complement(seq)

    return seq, sense


def process_V(imgt_annots, parsed_genes, accession):
    in_v_region = False
    parsing_gene = False
    l_part1s = []
    l_part2s = []
    v_heptamers = []
    v_nonamers = []


    row = {'accessions': accession, 'utr': None, 'l-part1': None, 'l-part2': None, 'v-heptamer': None, 'v-nonamer': None, 'v-region': None, 'sense': '', 'v-donor-splice': None, 'v-acceptor-splice': None}

    for line in imgt_annots:
        if 'FT   V-REGION' in line:
            in_v_region = True

            if parsing_gene:
                # We have a partially parsed truncated gene
                for el in ['utr', 'l-part1', 'l-part2', 'v-heptamer', 'v-nonamer', 'v-region']:
                    row[el] = str(row[el]) if row[el] is not None else ''
                parsed_genes.append(row)
                row = {'accessions': accession, 'utr': None, 'l-part1': None, 'l-part2': None, 'v-heptamer': None, 'v-nonamer': None, 'v-region': None}

            parsing_gene = True
            row['v-region'], row['sense'] = find_seq(line.replace('<', '').replace('>', ''))
        elif len(line) > 5 and line[5] != ' ' and in_v_region:
            in_v_region = False

        if 'FT   L-PART1' in line:
            row['l-part1'], _ = find_seq(line, 2) # include donor splice

        if 'FT   L-PART2' in line:
            row['l-part2'], _ = find_seq(line)

        if "FT   5'UTR" in line:
            row['utr'], _ = find_seq(line)

        if 'FT   V-HEPTAMER' in line:
            row['v-heptamer'], _ = find_seq(line)

        if 'FT   DONOR-SPLICE' in line:
            row['v-donor-splice'], _ = find_seq(line)

        if 'FT   ACCEPTOR-SPLICE' in line:
            row['v-acceptor-splice'], _ = find_seq(line)

        if 'FT   V-NONAMER' in line:
            row['v-nonamer'], _ = find_seq(line)
            l_part1s.append(row['l-part1'])
            l_part2s.append(row['l-part2'])
            v_heptamers.append(row['v-heptamer'])
            v_nonamers.append(row['v-nonamer'])
            for el in ['utr', 'l-part1', 'l-part2', 'v-heptamer', 'v-nonamer', 'v-region']:
                row[el] = str(row[el]) if row[el] is not None else ''
            parsed_genes.append(row)
            row = {'accessions': accession, 'utr': None, 'l-part1': None, 'l-part2': None, 'v-heptamer': None, 'v-nonamer': None, 'v-region': None}
            parsing_gene = False

        if in_v_region:
            if 'V-REGION' in line and '..' in line:
                reg = line.replace('FT   V-REGION', '')
                if 'complement' in reg:
                    reg = reg.replace('complement(', '').replace(')', '')
                reg = reg.split('..')
                row['start'] = reg[0].replace(' ', '')
                row['end'] = reg[1].replace('\n', '')
            elif 'IMGT_allele=' in line:
                row['allele'] = line.split('=')[1].replace('"', '').replace('\n', '')
            elif '/functional' in line:
                row['functional'] = 'functional'
            elif '/pseudo' in line:
                row['functional'] = 'pseudo'
            elif '/ORF' in line:
                row['functional'] = 'ORF'

    if parsing_gene:
        # We have a partially parsed truncated gene
        for el in ['utr', 'l-part1', 'l-part2', 'v-heptamer', 'v-nonamer', 'v-region']:
            row[el] = str(row[el]) if row[el] is not None else ''
        parsed_genes.append(row)

def process_D(imgt_annots, parsed_genes, accession):
    in_d_region = False
    parsing_gene = False

    d_3_heptamers = []
    d_3_nonamers = []
    d_3_spacers = []
    d_5_heptamers = []
    d_5_nonamers = []
    d_5_spacers = []

    row = {'accessions': accession, 'd-3-nonamer': None, 'd-3-spacer': None, 'd-3-heptamer': None, 'd-5-heptamer': None, 'd-5-spacer': None, 'd-5-nonamer': None}

    for line in imgt_annots:
        if 'FT   D-REGION' in line:
            if parsing_gene:
                for el in ['d-3-nonamer', 'd-3-spacer', 'd-3-heptamer', 'd-5-heptamer', 'd-5-spacer', 'd-5-nonamer']:
                    row[el] = str(row[el]) if row[el] is not None else ''
                parsed_genes.append(row)
                row = {'accessions': accession, 'd-3-nonamer': None, 'd-3-spacer': None, 'd-3-heptamer': None, 'd-5-heptamer': None, 'd-5-spacer': None,
                       'd-5-nonamer': None}
            parsing_gene = True
            in_d_region = True
        elif len(line) > 5 and line[5] != ' ' and in_d_region:
            in_d_region = False

        if "FT   5'D-HEPTAMER" in line:
            row['d-5-heptamer'], _ = find_seq(line)
        if "FT   5'D-SPACER" in line:
            row['d-5-spacer'], _ = find_seq(line)
        if "FT   5'D-NONAMER" in line:
            row['d-5-nonamer'], _ = find_seq(line)

        if "FT   3'D-HEPTAMER" in line:
            row['d-3-heptamer'], _ = find_seq(line)
        if "FT   3'D-SPACER" in line:
            row['d-3-spacer'], _ = find_seq(line)

        if "FT   3'D-NONAMER" in line:
            row['d-3-nonamer'], _ = find_seq(line)
            d_3_heptamers.append(row['d-3-heptamer'])
            d_3_nonamers.append(row['d-3-nonamer'])
            d_3_spacers.append(row['d-3-spacer'])
            d_5_heptamers.append(row['d-5-heptamer'])
            d_5_nonamers.append(row['d-5-nonamer'])
            d_5_spacers.append(row['d-5-spacer'])
            for el in ['d-3-nonamer', 'd-3-spacer', 'd-3-heptamer', 'd-5-heptamer', 'd-5-spacer', 'd-5-nonamer']:
                row[el] = str(row[el]) if row[el] is not None else ''
            parsed_genes.append(row)
            row = {'accessions': accession, 'd-3-nonamer': None, 'd-3-spacer': None, 'd-3-heptamer': None, 'd-5-heptamer': None, 'd-5-spacer': None, 'd-5-nonamer': None}
            parsing_gene = False

        if in_d_region:
            if 'FT   D-REGION' in line and '..' in line:
                reg = line.replace('FT   D-REGION', '')
                if 'complement' in reg:
                    reg = reg.replace('complement(', '').replace(')', '')
                reg = reg.split('..')
                row['start'] = reg[0].replace(' ', '')
                row['end'] = reg[1].replace('\n', '')
                row['d-region'], row['sense'] = find_seq(line)
            elif 'IMGT_allele=' in line:
                row['allele'] = line.split('=')[1].replace('"', '').replace('\n', '')
            elif '/functional' in line:
                row['functional'] = 'functional'
            elif '/pseudo' in line:
                row['functional'] = 'pseudo'
            elif '/ORF' in line:
                row['functional'] = 'ORF'

    if parsing_gene:
        for el in ['d-3-nonamer', 'd-3-spacer', 'd-3-heptamer', 'd-5-heptamer', 'd-5-spacer', 'd-5-nonamer']:
            row[el] = str(row[el]) if row[el] is not None else ''
        parsed_genes.append(row)


def process_J(imgt_annots, parsed_genes, accession):
    in_j_region = False
    in_j_gene = False
    j_heptamers = []
    j_nonamers = []
    j_spacers = []

    row = {'accessions': accession, 'start': None, 'end': None, 'allele': None, 'functional': None, 'j-nonamer': None, 'j-spacer': None, 'j-heptamer': None, 'j-region': None}

    for line in imgt_annots:
        if 'FT   J-GENE ' in line:
            in_j_gene = True
        if 'FT   J-REGION ' in line:
            in_j_region = True
        elif len(line) > 5 and line[5] != ' ' and in_j_region:
            in_j_region = False
            in_j_gene = False
            parsed_genes.append(row)
            row = {'accessions': accession, 'start': None, 'end': None, 'allele': None, 'sense': None, 'functional': None, 'j-nonamer': None, 'j-spacer': None, 'j-heptamer': None, 'j-region': None}

        if 'FT   J-HEPTAMER' in line:
            row['j-heptamer'], _ = find_seq(line)
            j_heptamers.append(row['j-heptamer'])
            j_nonamers.append(row['j-nonamer'])
            j_spacers.append(row['j-spacer'])
            for el in ['j-heptamer', 'j-nonamer', 'j-spacer']:
                row[el] = str(row[el]) if row[el] is not None else ''

        if 'FT   J-NONAMER' in line:
            row['j-nonamer'], _ = find_seq(line)

        if 'FT   J-SPACER' in line:
            row['j-spacer'], _ = find_seq(line)


        if in_j_region or in_j_gene:
            if 'FT   J-REGION ' in line and '..' in line:
                reg = line.replace('FT   J-REGION ', '')
                if 'complement' in reg:
                    reg = reg.replace('complement(', '').replace(')', '')
                reg = reg.split('..')
                row['start'] = reg[0].replace(' ', '')
                row['end'] = reg[1].replace('\n', '')
                row['j-region'], row['sense'] = find_seq(line)
            elif 'IMGT_allele=' in line:
                row['allele'] = line.split('=')[1].replace('"', '').replace('\n', '')
            elif '/functional' in line:
                row['functional'] = 'functional'
            elif '/pseudo' in line:
                row['functional'] = 'pseudo'
            elif '/ORF' in line:
                row['functional'] = 'ORF'


def main():
    global assembly
    sequences = {}
    args = get_parser().parse_args()

    if args.locus not in ['IGH', 'IGK', 'IGL']:
        print('error: locus must me IGH, IGK or IGL')
        exit(0)

    if 'http' in args.imgt_url:
        with urllib.request.urlopen(args.imgt_url) as fi:
            imgt_text = html.unescape(fi.read().decode('utf-8'))

            if args.save_download:
                with open(args.save_download, 'w') as fo:
                    fo.write(imgt_text)
    else:
        with open(args.imgt_url, 'r') as fi:
            imgt_text = fi.read()

    all_annotations = []
    parsed_genes = []
    annotations = []
    sequence = ""
    reading = 'preamble'
    accessions = ''
    fieldnames = None

    for line in imgt_text.split('\n'):
        if reading == 'preamble':
            annotations.append(line)
            if len(line) > 1 and line[0:2] == 'FT':
                reading = 'annotations'
            if len(line) > 5 and line[0:2] == 'AC':
                accessions = line[5:]
                print(f'Processing {accessions}')

        elif reading == 'annotations':
            annotations.append(line)
            if len(line) < 2 or line[0:2] != 'FT':
                reading = 'postamble'

        elif reading == 'postamble':
            if len(line) > 1 and line[0:2] == 'SQ':
                reading = 'sequence'
            else:
                annotations.append(line)

        elif reading == 'sequence':
            if line[0:5] == '     ':
                sequence += line[5:70].replace(' ', '')
            else:
                assembly = sequence

                if args.save_sequence and accessions:
                    sequences[accessions] = assembly

                process_V(annotations, parsed_genes, accessions)

                if args.locus in ['IGH', 'TRB', 'TRD']:
                    process_D(annotations, parsed_genes, accessions)

                process_J(annotations, parsed_genes, accessions)

                fieldnames = [
                   'accessions', 'allele', 'functional',  'start', 'end', 'sense',
                   'utr', 'l-part1', 'l-part2', 'v-region',  'v-nonamer', 'v-heptamer', 'v-donor-splice', 'v-acceptor-splice',
                   'd-5-heptamer', 'd-5-spacer', 'd-5-nonamer', 'd-region','d-3-heptamer', 'd-3-spacer', 'd-3-nonamer',
                   'j-region', 'j-heptamer', 'j-spacer', 'j-nonamer',
                ]

                sequence = ""
                reading = 'preamble'
                accessions = ''
                all_annotations.extend(annotations)
                annotations = []

    if args.save_imgt_annots:
        with open(args.save_imgt_annots, 'w') as fo:
            fo.write('\n'.join(all_annotations))

    if args.save_sequence:
        simple.write_fasta(args.save_sequence, sequences)

    if fieldnames:
        with open(args.outfile, 'w', newline='') as fo:
            writer = csv.DictWriter(fo, fieldnames=fieldnames, restval='')
            writer.writeheader()
            for parsed_gene in parsed_genes:
                writer.writerow(parsed_gene)
    else:
        print("No annotation records found.")


if __name__ == "__main__":
    main()
